pub mod models;
pub mod schema;
pub mod utils;

use diesel::prelude::*;
use dotenvy::dotenv;
use schema::simulation_behaviors;
use schema::simulation_behaviors::dsl::*;
use std::env;
use utils::mobility_data::MobilityBehavior;

use models::{NewSimulationBehavior, SimulationBehavior};

pub fn establish_connection() -> SqliteConnection {
    dotenv().ok();

    let database_url = env::var("DATABASE_URL").unwrap_or_else(|_| "./injector.db".into());
    // .expect("DATABASE_URL must be set");
    SqliteConnection::establish(&database_url)
        .unwrap_or_else(|_| panic!("Error connecting to {}", database_url))
}

pub fn create_behavior(
    conn: &mut SqliteConnection,
    record: &MobilityBehavior,
) -> SimulationBehavior {
    let new_behavior = NewSimulationBehavior {
        dataset_id: &(record.dataset_id as i32),
        relative_start_time: &(record.relative_start_time as i32),
        quantity: &record.quantity.to_string(),
        mobility_record: &record.mobility_record.to_bytes().to_vec(),
    };

    diesel::insert_into(simulation_behaviors::table)
        .values(&new_behavior)
        .returning(SimulationBehavior::as_returning())
        .get_result(conn)
        .expect("Error saving new simulation behavior")
}

pub fn delete_behaviors_for_dataset(
    conn: &mut SqliteConnection,
    dataset_id_to_delete: i32,
) -> Result<usize, diesel::result::Error> {
    diesel::delete(simulation_behaviors.filter(dataset_id.eq(dataset_id_to_delete))).execute(conn)
}

pub fn delete_behavior(
    conn: &mut SqliteConnection,
    id_to_delete: i32,
) -> Result<usize, diesel::result::Error> {
    diesel::delete(simulation_behaviors.filter(id.eq(id_to_delete))).execute(conn)
}

pub fn find_behavior_by_id(
    conn: &mut SqliteConnection,
    record_id: i32,
) -> Result<SimulationBehavior, diesel::result::Error> {
    simulation_behaviors.find(record_id).first(conn)
}

pub fn find_next_behavior(
    conn: &mut SqliteConnection,
) -> Result<SimulationBehavior, diesel::result::Error> {
    simulation_behaviors.first(conn)
}

// overflow evaluating the requirement `(&_, &_, &_, ..., ..., ...): diesel::Insertable<...>`
// consider increasing the recursion limit by adding a `#![recursion_limit = "256"]` attribute to your crate (`testnet_injector`)
// ---
// pub fn create_behaviors(
//     conn: &mut SqliteConnection,
//     mobility_behaviors: &'_ [MobilityBehavior; 10],
// ) -> QueryResult<usize> {
//     let new_behaviors = mobility_behaviors
//         .iter()
//         .map(|mb| NewSimulationBehavior {
//             dataset_id: &(mb.dataset_id as i32),
//             relative_start_time: &(mb.relative_start_time as i32),
//             quantity: &mb.quantity.to_string(),
//             mobility_record: &mb.mobility_record.to_bytes().to_vec(),
//         })
//         .collect();

//     diesel::insert_into(simulation_behaviors::table)
//         .values(&new_behaviors)
//         // .returning(Behavior::as_returning())
//         .execute(conn)
// }
