use diesel::r2d2::ConnectionManager;
use diesel::{SqliteConnection};
use dotenvy::dotenv;
use ecomobicoin_jsonrpc::types::{Behavior, Block};
use ethereum_types::U64;
use ethers::{prelude::*, signers::coins_bip39::English};
use futures::executor::block_on;
use serde::{Deserialize};
use testnet_injector::utils::miner::get_miner_behavior;
use std::env;
use std::{process, time::Duration};
use testnet_injector::{
    create_behavior, find_behavior_by_id,
};
use testnet_injector::{
    utils::{mobility_data::*, txbx},
};
use tracing::{debug, error, info, Level};

use ::clap::{Args, Parser, Subcommand};

#[derive(Parser)]
#[command(author, version)]
#[command(
    about = "stringer - a simple CLI to transform and inspect strings",
    long_about = "stringer is a super fancy CLI (kidding)

One can use stringer to modify or inspect strings straight from the terminal"
)]
struct Cli {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    /// Send txs and/or bxs using default content
    Send(SendParams),
    /// Simulate behaviors from dataset in csv and send them to node
    SimulateBxFromCsv(SimulateBxFromCsv),
    /// Mine a block
    Mine(MineParams),
}
// TODO move .env to cli params

#[derive(Args)]
struct SendParams {
    /// Number of transactions to send
    txs: u64,
    /// Number of behaviors to send
    bxs: u64,
}

#[derive(Args)]
struct MineParams {
}

#[derive(Args)]
struct SimulateBxFromCsv {
    /// Path to CSV file
    file: String,
    /// Total duration of simulation in seconds
    total_simulation_duration: u64,
}

#[derive(Deserialize)]
struct BlockFeed {
    txs: usize,
    bxs: usize,
}

async fn init() -> (u64, String) {
    // construct a subscriber that prints formatted traces to stdout
    let subscriber = tracing_subscriber::FmtSubscriber::builder()
        .with_max_level(Level::INFO)
        .finish();
    // use that subscriber to process traces emitted after this point
    tracing::subscriber::set_global_default(subscriber)
        .expect("setting default tracing subscriber failed");

    dotenv().expect(".env file not found");
    // for (key, value) in std::env::vars() {
    //     println!("{key}: {value}");
    // }
    let rpc_url: String =
        std::env::var("RPC_URL").unwrap_or_else(|_| "http://127.0.0.1:8545".into());
    let provider = Provider::<Http>::try_from(rpc_url.clone()).unwrap();

    let chain_id: u64 = std::env::var("CHAIN_ID")
        .unwrap_or_else(|_| String::from("63"))
        .parse::<u64>()
        .unwrap();

    let block_number: U64 = provider.get_block_number().await.unwrap_or_else(|_| {
        panic!(
            "Failed to get block number from {:?}, cannot go further",
            rpc_url
        )
    });
    info!(
        "Connected to chain id: {:?}, rpc: {:?}, latest block number: {:?}.",
        chain_id, rpc_url, block_number
    );

    (chain_id, rpc_url)
}

#[tokio::main]
async fn main() -> Result<(), anyhow::Error> {
    let cli = Cli::parse();

    let (chain_id, rpc_url) = init().await;

    match &cli.command {
        Some(Commands::Send(params)) => {
            info!("Sending {:?} tx and {:?} bx ...", params.txs, params.bxs);
            
            let max = std::cmp::max(params.txs, params.bxs);

            for i in 0..max {
                let mnemonic = std::env::var("MNEMONIC").unwrap_or("abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about".to_string());
                let wallet = MnemonicBuilder::<English>::default()
                .phrase(PathOrString::String(mnemonic.clone()))
                .index(i as u32)
                .unwrap()
                .build()
                .unwrap()
                .with_chain_id(chain_id);
                let provider = Provider::<Http>::try_from(rpc_url.clone()).unwrap();
                let client = SignerMiddleware::new(provider.clone(), wallet.clone());


                if i < params.bxs {
                    let bx_hash = txbx::send_default_bx(
                        client.clone(),
                        chain_id,
                    ).await;
                    info!("Sent default bx {:?}", bx_hash);
                }
                if i < params.txs {
                    let tx_hash = txbx::send_default_tx(
                        client.clone(),
                        chain_id,
                    ).await;
                    info!("Sent default tx {:?}", tx_hash);
                }
            }
        }
        Some(Commands::SimulateBxFromCsv(params)) => {
            // println!("Processing {:?}", params.file);
            // csv::Reader::from_path(path)
            let records = read_csv(csv::Reader::from_path(params.file.clone())?);
            if let Err(err) = records {
                error!("error running example: {}", err);
                process::exit(1);
            }
            let records = records.unwrap();

            let database_url = env::var("DATABASE_URL").unwrap_or_else(|_| "./injector.db".into());
            // .expect("DATABASE_URL must be set");
            let manager = ConnectionManager::<SqliteConnection>::new(&database_url);
            let pool = r2d2::Pool::builder()
                .build(manager)
                .expect("Failed to create pool.");

            let connection = &mut pool.get().unwrap();

            let mobility_records = generate_behaviors_from_records(records.clone());
            let dataset_id = mobility_records[0].dataset_id;

            let dataset_max_relative_time = mobility_records
                .iter()
                .max_by(|x, y| x.relative_start_time.cmp(&y.relative_start_time))
                .unwrap()
                .relative_start_time;

            let total_simulation_duration = if params.total_simulation_duration > 0 {
                params.total_simulation_duration
            } else {
                dataset_max_relative_time as u64
            };

            info!(
                "Simulation of {:?} records for {:?}secs (max time {:?})",
                mobility_records.len(),
                total_simulation_duration,
                Duration::from_millis(dataset_max_relative_time as u64)
            );

            // Save mobility records in DB
            let mut mobility_records_ids = Vec::with_capacity(mobility_records.len());
            let mut mobility_records_times = Vec::with_capacity(mobility_records.len());
            for record in mobility_records.iter() {
                //TODO insert multiples but rust error
                let insert_result = create_behavior(connection, record);
                mobility_records_ids.push(insert_result.id);
                mobility_records_times.push(compute_stretched_relative_start_time(
                    record.relative_start_time as u64,
                    total_simulation_duration,
                    dataset_max_relative_time as u64,
                ));
            }
            debug!(
                "Saved {:?} mobility records in DB for dataset_id {:?}",
                mobility_records.len(),
                mobility_records[0].dataset_id
            );

            // ----------------
            // Sending queue
            // ----------------
            let mut handles = Vec::with_capacity(mobility_records_ids.len());

            for i in 0..mobility_records_ids.len() {
                let pool = pool.clone();
                let simulation_behavior_id = mobility_records_ids[i].clone();
                let relative_start_duration = mobility_records_times[i].clone();
                let rpc_url = rpc_url.clone();
                let chain_id = chain_id.clone();

                handles.push(tokio::spawn({
                    async move {
                            tokio::time::sleep(relative_start_duration).await;
                            
                            let provider = Provider::<Http>::try_from(rpc_url.clone()).unwrap();

                            let conn = &mut pool.get().unwrap();


                            let behavior = find_behavior_by_id(conn,simulation_behavior_id).unwrap();
                            debug!("Waited {:?} for behavior {:?}",relative_start_duration,behavior.id);
                            // Wallet is derived from mnemomic with user_id to ensure each user have the same wallet for the simulation.
                            let mnemonic = std::env::var("MNEMONIC").unwrap_or("abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about".to_string());
                            let wallet = MnemonicBuilder::<English>::default()
                                .phrase(PathOrString::String(mnemonic.clone()))
                                .index(simulation_behavior_id as u32)
                                .unwrap()
                                .build()
                                .unwrap()
                                .with_chain_id(chain_id);

                            info!(
                                "Sending bx {:?}, qty: {:?},user {:?} with wallet {:?} ",
                                behavior.id,
                                behavior.quantity,
                                simulation_behavior_id,
                                wallet.address()
                            );

                            let client = SignerMiddleware::new(provider.clone(), wallet.clone());

                            let bx_hash = txbx::send_behavior_transaction(
                                client.clone(),
                                chain_id,
                                behavior.relative_start_time as u64,
                                U256::from_dec_str(&behavior.quantity).unwrap(),
                                Bytes::default(), // Bytes::from_static(&behavior.mobility_record),
                            ).await;
                            info!("Sent bx({:?}): {:?}",behavior.id,bx_hash);
                            let tx_hash = txbx::send_default_tx(client, chain_id).await;
                            info!("Sent tx({:?}): {:?}",behavior.id,tx_hash);
                    }
                        }))
            }

            info!("Scheduling done, waiting for bx & tx threads...");

            // Wait for all of them to complete.
            for handle in handles {
                block_on(handle).unwrap();
            }
        },
        Some(Commands::Mine(params)) => {
            
            // Wallet is derived from mnemomic with user_id to ensure each user have the same wallet for the simulation.
            let miner_signing_key = std::env::var("MINER_SIGNING_KEY").expect("Must specify a miner private key");
            let wallet: LocalWallet = miner_signing_key.parse::<LocalWallet>().unwrap();
            let miner_addr = wallet.address();
            info!("Mining with 0x{:02x} ...",miner_addr);
            
            let behavior = get_miner_behavior(wallet, chain_id, miner_addr);
            // SEND IT
            let provider = Provider::<Http>::try_from(rpc_url.clone()).unwrap();
            debug!("Miner behavior: {:?}",serde_json::to_string(&behavior.clone()));
            let block = provider.request::<[Behavior;1], Block>("emc_mine", [behavior]).await.unwrap();

            debug!("Mining response: {:?}", block);
            info!("Mined block {:?}", block.number.unwrap())
        },
        None => {}
    }
    Ok(())
}
