use std::{error::Error, ops::Sub, time::Duration};

use csv::Reader;
use ethers::{
    prelude::rand::{thread_rng, Rng},
    types::Bytes,
};
use primitive_types::U256;
use serde::Deserialize;
use time::OffsetDateTime;

/// Mobility Record from Public Transportation
/// ---
/// Created from data in `data/dataset-extract.csv`
///
/// CSV header:
/// id user_id network service label created_at from_coord from_label from_datetime from_latitude from_longitude to_coord to_label to_datetime to_latitude to_longitude
#[derive(Debug, Deserialize, Clone)]
pub struct MobilityRecord {
    pub id: u64,
    pub user_id: u64,
    pub network: String,
    pub service: String,
    pub label: String,
    // `YYYY-MM-DD HH:mm:ss.SSSSSS` rfc3339
    #[serde(with = "time::serde::rfc3339")]
    pub created_at: time::OffsetDateTime,
    /// Departure coordinates in hex PostGIS EWKB (Extended Well-Known Binary)
    pub from_coord: Option<String>,
    pub from_label: Option<String>,
    // either `YYYY-MM-DD HH:mm:ss` or `YYYY-MM-DD HH:mm:ss.SSS` rfc3339
    #[serde(with = "time::serde::rfc3339::option")]
    pub from_datetime: Option<time::OffsetDateTime>,
    pub from_latitude: Option<f64>,
    pub from_longitude: Option<f64>,
    /// Arrival coordinates in hex PostGIS EWKB (Extended Well-Known Binary)
    pub to_coord: Option<String>,
    pub to_label: Option<String>,
    // `YYYY-MM-DD HH:mm:ss` rfc3339
    #[serde(with = "time::serde::rfc3339::option")]
    pub to_datetime: Option<time::OffsetDateTime>,
    pub to_latitude: Option<f64>,
    pub to_longitude: Option<f64>,
}

impl MobilityRecord {
    unsafe fn any_as_u8_slice<T: Sized>(p: &T) -> &[u8] {
        ::core::slice::from_raw_parts((p as *const T) as *const u8, ::core::mem::size_of::<T>())
    }
    pub fn to_bytes(&self) -> Bytes {
        let bytes: &[u8] = unsafe {
            ::core::slice::from_raw_parts(
                (self as *const MobilityRecord) as *const u8,
                ::core::mem::size_of::<MobilityRecord>(),
            )
        };

        Bytes::from_static(bytes)
    }
}

pub fn read_csv<R: std::io::Read>(
    mut reader: Reader<R>,
) -> Result<Vec<MobilityRecord>, Box<dyn Error>> {
    let mut records = vec![];
    for result in reader.deserialize() {
        let record: MobilityRecord = result?;
        // println!("{:?}", record);
        records.push(record)
    }
    Ok(records)
}

/// Mobility Behavior
/// ---
///
/// Including a MobilityRecord from CSV extraction
pub struct MobilityBehavior {
    pub id: u64,
    pub dataset_id: u64,
    // in milliseconds
    pub relative_start_time: i128,
    pub quantity: U256,
    pub mobility_record: MobilityRecord,
}

pub fn find_earliest_mobility_record(records: &Vec<MobilityRecord>) -> OffsetDateTime {
    records
        .iter()
        .min_by(|x, y| x.created_at.cmp(&y.created_at))
        .unwrap()
        .created_at
}

pub fn find_latest_mobility_record(records: &Vec<MobilityRecord>) -> OffsetDateTime {
    records
        .iter()
        .max_by(|x, y| x.created_at.cmp(&y.created_at))
        .unwrap()
        .created_at
}

pub fn generate_behaviors_from_records(records: Vec<MobilityRecord>) -> Vec<MobilityBehavior> {
    let earliest_datetime = find_earliest_mobility_record(&records);

    let dataset_id = thread_rng().gen::<u64>();

    records
        .iter()
        .map(|r| MobilityBehavior {
            id: 0,
            dataset_id,
            relative_start_time: r.created_at.sub(earliest_datetime).whole_milliseconds(),
            quantity: U256::from(1337),
            mobility_record: r.clone(),
        })
        .collect()
}

/// Strech behavior start time for simulation
pub fn compute_stretched_relative_start_time(
    relative_start_time: u64,
    total_simulation_duration: u64,
    dataset_max_relative_time: u64,
) -> Duration {
    Duration::from_millis(
        (relative_start_time * total_simulation_duration * 1000) / dataset_max_relative_time,
    )
}

#[cfg(test)]
mod tests {
    use time::OffsetDateTime;

    use super::*;

    #[test]
    fn read_csv_should_work() {
        // ARRANGE
        let example_data = "\
        id,user_id,network,service,label,created_at,from_coord,from_label,from_datetime,to_coord,to_label,to_datetime,from_latitude,from_longitude,to_latitude,to_longitude
        10653,79,mulhouse,PUBLIC_TRANSPORT,Abo,2023-07-02T19:02:04.987621Z,DEADC0DE,TEST,2023-07-02T21:02:04.000Z,DEADC0DE,TEST,2023-07-02T21:02:04.000Z,0.0,0.0,0.0,0.0
        ";
        // let mut rdr = Reader::from_reader(example_data.as_bytes());
        let mut rdr = Reader::from_path("./data/dataset-extract.csv").unwrap();
        // ACT
        let result = read_csv(rdr).unwrap();
        // ASSERT
        assert_eq!(6, result.len())
    }

    #[test]
    fn to_bytes_should_work() {
        // ARRANGE
        let mobility_record = MobilityRecord {
            id: 1337,
            user_id: 1,
            network: String::from("mulhouse"),
            service: String::from("PUBLIC_TRANSPORT"),
            label: String::from("BLABLABLA"),
            created_at: OffsetDateTime::from_unix_timestamp(1699353308).unwrap(),
            from_datetime: None,
            from_coord: None,
            from_label: None,
            from_latitude: None,
            from_longitude: None,
            to_datetime: None,
            to_coord: None,
            to_label: None,
            to_latitude: None,
            to_longitude: None,
        };

        // ACT
        let result = mobility_record.to_bytes();
        // ASSERT

        // FIXME Binary data change between tests o
        assert!(!result.to_string().is_empty());
        // assert_eq!("0x", result.to_string());
    }

    #[test]
    fn find_earliest_mobility_record_should_work() {
        // ARRANGE
        let mut rdr = Reader::from_path("./data/dataset-extract.csv").unwrap();
        let records = read_csv(rdr).unwrap();

        // ACT
        let earliest_datetime = find_earliest_mobility_record(&records);

        // ASSERT
        assert_eq!(
            1685570991226773000,
            earliest_datetime.unix_timestamp_nanos()
        );
    }

    #[test]
    fn find_latest_mobility_record_should_work() {
        // ARRANGE
        let mut rdr = Reader::from_path("./data/dataset-extract.csv").unwrap();
        let records = read_csv(rdr).unwrap();

        // ACT
        let latest_datetime = find_latest_mobility_record(&records);

        // ASSERT
        assert_eq!(1688324524987621000, latest_datetime.unix_timestamp_nanos());
    }

    #[test]
    fn create_behaviors_from_records_should_work() {
        // ARRANGE
        let mut rdr = Reader::from_path("./data/dataset-extract.csv").unwrap();
        let records = read_csv(rdr).unwrap();

        // ACT
        let behaviors = generate_behaviors_from_records(records.clone());
        // ASSERT
        assert_eq!(records.len(), behaviors.len());
    }

    #[test]
    fn compute_stretched_relative_start_time_should_work_0() {
        // ARRANGE

        // ACT
        let result = compute_stretched_relative_start_time(0, 10, 1000);

        // ASSERT
        assert_eq!(Duration::from_secs(0), result);
    }

    #[test]
    fn compute_stretched_relative_start_time_should_work() {
        // ARRANGE
        // 1000 ms
        let relative_start_time = 1000;
        // 10 secs
        let total_simulation_duration = 10;
        // 2 sec
        let dataset_max_relative_time = 2000;

        // ACT
        let result = compute_stretched_relative_start_time(
            relative_start_time,
            total_simulation_duration,
            dataset_max_relative_time,
        );
        // ASSERT
        assert_eq!(Duration::from_secs(5), result);
    }
    #[test]
    fn compute_stretched_relative_start_time_should_work_max() {
        // ARRANGE
        // 1000 ms
        let relative_start_time = 2000;
        // 10 secs
        let total_simulation_duration = 10;
        // 2 sec
        let dataset_max_relative_time = 2000;

        // ACT
        let result = compute_stretched_relative_start_time(
            relative_start_time,
            total_simulation_duration,
            dataset_max_relative_time,
        );
        // ASSERT
        assert_eq!(Duration::from_secs(10), result);
    }
}
